from ._base_distribution import (
    AbstractDistribution as AbstractDistribution,
    AbstractMarginalDistribution as AbstractMarginalDistribution,
)
from ._gaussian_distributions import (
    IndependentGaussianFourierModes as IndependentGaussianFourierModes,
)
